
package jst.symtab;
import jst.app.JSTMessage;
import jst.parsetree.*;
import jst.symtab.link.*;
import parser.parsetree.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.Modifier;
import jst.symtab.visitor.ModelVisitor;

public abstract class ClassType extends UserType {

    /**
     * @supplierCardinality 0..1 
     * @supplierRole superclass
     */
	protected UserTypeLookup maybeSuperclassRef; // Will be null after resolved.
	protected ClassTypeLookup superclassRef; // Guaranteed non null after resolved.

    /** 
     * @supplierCardinality 0..*
     * @supplierRole subclasses*/
	protected Set<ClassType> subclasses;

    /**
     * @supplierCardinality 0..*
     * @supplierRole interfaceRefs*/
	protected Set<InterfaceTypeLookup> interfaceRefs;

    /**
     *@link aggregation
     *      @associates <{ConstructorDecl}>
     * @supplierCardinality 0..*
     * @supplierRole constructors
     */
    protected Set<ConstructorDecl> constructors;

    /**
     *@link aggregation
     *      @associates <{Block}>
     * @supplierCardinality 0..*
     * @supplierRole initializerBlocks
     */
    protected List<Block> initializerBlocks;

	protected ClassType(SourceFile ownerSourceFile, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerSourceFile, theSimpleName, theSource);
    }

	protected ClassType(PackageDecl ownerPackage, String theSimpleName, JstParseTreeBranch theSource) {
		super(ownerPackage, theSimpleName, theSource);
    }

	protected ClassType(UserType ownerUserType, String theSimpleName, JstParseTreeBranch theSource) {
		super(ownerUserType, theSimpleName, theSource);
    }

	protected ClassType(Block ownerBlock, String theSimpleName, JstParseTreeBranch theSource) {
		super(ownerBlock, theSimpleName, theSource);
    }

	protected ClassType(FieldInitializer ownerFieldInitializer, String theSimpleName, JstParseTreeBranch theSource) {
		super(ownerFieldInitializer, theSimpleName, theSource);
    }

	// Constructor used by ParameterizedClassType subclass
	// Needed to set owner without adding the ParameterizedClassType to the owner's types collection.
	protected ClassType(GenericClassType theOwner, String theSimpleName, JstParseTreeBranch theSource) {
    	super(theOwner, theSimpleName, theSource);
    }

	protected ClassType(SourceFile newOwnerSourceFile, ClassType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerSourceFile, that, swaps);

		init(that, swaps);
	}

	protected ClassType(PackageDecl newOwnerPackage, ClassType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerPackage, that, swaps);

		init(that, swaps);
	}

	protected ClassType(UserType newOwnerUserType, ClassType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerUserType, that, swaps);

		init(that, swaps);
	}

	protected ClassType(Block newOwnerBlock, ClassType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerBlock, that, swaps);

		init(that, swaps);
	}

	protected ClassType(FieldInitializer newOwnerFieldInitializer, ClassType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerFieldInitializer, that, swaps);

		init(that, swaps);
	}

	// Tricksy constructor for ParameterizedClassType subclass.
	protected ClassType(GenericClassType newOwner,  ClassType that, Map<TypeParameter, ActualType> swaps) {
    	super(newOwner, that, swaps);

		// Subclass adds the Parameterized usertype to the Generic owner.

		init(that, swaps);
	}

    protected void init() {
		super.init();

        subclasses = new HashSet<ClassType>();
        interfaceRefs = new HashSet<InterfaceTypeLookup>();
        constructors = new HashSet<ConstructorDecl>();
        initializerBlocks = new ArrayList<Block>();
	}

	protected void init(ClassType that, Map<TypeParameter, ActualType> swaps) {
		if (that.maybeSuperclassRef != null)
			maybeSuperclassRef = that.maybeSuperclassRef.copy(this);

		if (that.superclassRef != null)
			superclassRef = that.superclassRef.copy(this);

		for (InterfaceTypeLookup thatFace : that.interfaceRefs)
			thatFace.copy(this);
		
		for (ConstructorDecl thatCon : that.constructors)
			thatCon.copy(this, swaps);
		
		for (Block thatInit : that.initializerBlocks)
			thatInit.copy(this, swaps);
	}

	public ClassType asClass() {
		return this;
	}

	public void setSuperclassRef(UserTypeLookup superRef) {
    	maybeSuperclassRef = superRef;
    }

	public void setSuperclassRef(ClassTypeLookup superRef) {
    	superclassRef = superRef;
    }

	public boolean resolveSuperclass(List messages) {

		// Eww! Anonymous classes inherit from either a class or an interface &
		// we can't tell which until types are looked up.
		if (maybeSuperclassRef != null) {

			assert superclassRef == null;

			if (!maybeSuperclassRef.resolve()) {
				messages.add(maybeSuperclassRef.getMessage());
				return false;
			}

			if (maybeSuperclassRef.getUserType().asClass() != null) {
				superclassRef = new ClassTypeLookup(maybeSuperclassRef);
			}
			else {
				int line = maybeSuperclassRef.getLine();
				int col  = maybeSuperclassRef.getColumn();

				ScopeLink link = new KnownDecl(this);
				link = new PackageLookup(link, new StringIdentifier("java", line, col));
				link = new PackageLookup(link, new StringIdentifier("lang", line, col));
				superclassRef = new ClassTypeLookup(link, new StringIdentifier("Object", line, col));

				addSuperInterfaceRef(new InterfaceTypeLookup(maybeSuperclassRef));
			}

			maybeSuperclassRef = null;

			// Fall through to resolve superclass normally.
		}

		assert superclassRef != null;

		if (!superclassRef.resolve()) {
			messages.add(superclassRef.getMessage());
			return false;
		}

		getSuperclass().addSubclass(this);
		return true;
	}

	public ClassTypeLookup getSuperclassRef() {
		return superclassRef;
	}

	public ClassType getSuperclass() {
		if (superclassRef == null) // Must be trying to work out a maybeSuperclassRef still
			return null;

    	return superclassRef.getClassType();
    }

	protected void addSubclass(ClassType subclass) {
		subclasses.add(subclass);
	}
	 
	public Set<ClassType> getSubclasses() {
		return subclasses;
	}

	public void addSuperInterfaceRef(InterfaceTypeLookup interfaceLookup) {
    	interfaceRefs.add(interfaceLookup);
    }

	public boolean resolveInterfaces(List messages) {

		boolean ok = true;

    	for(InterfaceTypeLookup interfaceRef : interfaceRefs){

			if (!interfaceRef.resolve()) {
				ok = false;
				messages.add(interfaceRef.getMessage());
			}
			else
				interfaceRef.getInterfaceType().addImplementedBy(this);
		}

		return ok;
	}

	public Set<InterfaceType> getInterfaces() {
		Set<InterfaceType> interfaces = new HashSet<InterfaceType>();

		for(InterfaceTypeLookup ref : interfaceRefs)
			if (ref.getInterfaceType() != null)
				interfaces.add(ref.getInterfaceType());

		return interfaces;
	}

	public void addConstructor(ConstructorDecl constructor) {
		constructors.add(constructor);
	}

	public Set getConstructors() {
		return constructors;
	}

	public void addInitializerBlock(Block initializer) {
		initializerBlocks.add(initializer);
	}

	public List getInitializerBlocks() {
		return initializerBlocks;
	}

	public boolean isSubtypeOf(ReferenceType parent) {
		if (parent == this)
			return true;
			
		ClassType superclass = getSuperclass();
		if (superclass != null && superclass.isSubtypeOf(parent))
			return true;

		Iterator ifaceIter = getInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			if (iface.isSubtypeOf(parent))
            	return true;
        }

		return false;
    }

	public VariableDecl lookupVariable(boolean widen, String variableName, ReferenceType looker) {
    	FieldDecl field = (FieldDecl) fields.get(variableName);

        if (field != null)
			return field.accessible(looker)? field : null;

		VariableDecl variable = null;
		
		ClassType superclass = getSuperclass();
        if (superclass != null)
        	variable = superclass.lookupVariable(false, variableName, looker);

        if (variable != null)
			return variable;

		Iterator ifaceIter = getInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			variable = iface.lookupVariable(false, variableName, looker);
			if (variable != null)
				return variable;
		}

		return widen? getOwnerScope().lookupVariable(true, variableName, looker) : null;
	}

	public ConstructorDecl lookupConstructor(List params, ReferenceType looker) {
    	// Constructors are not inherited.

		if (constructors.isEmpty()) {
        	if (params.isEmpty()) // Make a default constructor.
            	return defaultConstructor();

			return null;
        }

		return (ConstructorDecl) chooseOperation(constructors, params, looker);
    }

	public TypeDecl lookupType(boolean widen, String typeName, ReferenceType looker) {
    	// Tricky: The name of a class needs to appear in its own scope.  This is necessary
        // in cases such as import statements, where the import must bring the class name into scope.
        // However, we don't want the class to contain itself, so just check for a lookup of the
        // class name.
        if (typeName.equals(simpleName))
        	return this;

		if (typeName.equals("this"))
			return this;

		if (typeName.equals("super"))
			return getSuperclass();

    	TypeDecl result = (TypeDecl) types.get(typeName);
        if (result != null)
			return result.accessible(looker)? result : null;

		ClassType superclass = getSuperclass();
        if (superclass != null)
			result = superclass.lookupType(false, typeName, looker);

        if (result != null)
			return result;

		Iterator ifaceIter = getInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			result = iface.lookupType(false, typeName, looker);
			if (result != null)
				return result;
		}

		return widen? getOwnerScope().lookupType(true, typeName, looker) : null;
   }

	protected Set findInheritedMethodsNamed(String methodName) {
    	Set result = new HashSet();

		ClassType superclass = getSuperclass();
    	if (superclass != null)
    	   result.addAll(superclass.findAllMethodsNamed(methodName));

		Iterator ifaceIter = getInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			result.addAll(iface.findAllMethodsNamed(methodName));
        }

        /* NYET! Only look at outer class when doing a named symbol table lookup.
           Outer class methods should not be found within the inner class.
           When searching from within the inner class, an owner scope will contain the goal.

		// Try outer class (not really "inherited")
		// If this is an inner class, & we know it is static, we don't need to search outer.
        UserType outer = getOwnerUserType();
        if (outer != null &&  Modifier.isStatic(modifier))
        	result.addAll(outer.findAllMethodsNamed(methodName));
        */

		return result;
    }

	protected ConstructorDecl defaultConstructor() {
        ConstructorDecl result = new ConstructorDecl(this, null);
        result.setModifier(Modifier.PUBLIC);
        return result;
    }

    public void accept(ModelVisitor visitor) {
        visitor.visitClassType(this);
    }
}



