package parser.parsetree;

import parser.parsetree.visitor.*;
import parser.grammar.*;

import java.util.*;

/**
* An Abstract Syntax Tree list of parse tree nodes.  BNF allows lists only through
* recursive productions.  This can make parse trees ugly to traverse.  An AST
* builder can replace the left-recursive production idiom with a flattened ASTList.
* <p>
* An instance of ASTList has an arbitrary number of kids.  This replaces the
* (arbitrarily deep) recursive tree.
* <p>
* The left-recursion idiom looks like:
*   X ::= X Y  (Production X_1)
*      |  Y    (Production X_2)
*
* An example parse tree is:
*       X   (An X_1)
*     /  \
*     X  |  (An X_1)
*   /  \ |
*   X  | |  (An X_1)
*  / \ | |
*  X | | |  (An X_2)
*  | | | |
*  Y Y Y Y
* 
* A single ASTList directly contains all four Y's.  The Symbol type of the list
* is the lhs of both productions, i.e. X.
* It is the responsibility of the client to recognise that trees using
* the X symbol cannot be traversed in the usual way: the kids are in a flat list.
* <p>
* This is a convenience feature.  In the presence of ambiguities, e.g. Y's that
* span different sections of the input, it stops being convenient.  Consequently,
* no attempt is made to support automatic AST construction for grammars that
* have ambiguous lists.
*/
public class ASTList extends SyntaxTreeArrayBranch {

    protected Nonterminal nonterminal;

    /** Construct an ASTList. */
	public ASTList(Production productionType, SyntaxTree firstKid) {
		super(makeKidArray(firstKid));

        nonterminal = productionType.getLhs();

    }

	protected static SyntaxTree[] makeKidArray(SyntaxTree firstKid) {
		SyntaxTree[] result = new SyntaxTree[1]; 
        result[0] = firstKid;
		return result;
	}

    /** Construct an ASTList that extends an existing list. */
	public ASTList(Production productionType, ASTList existing, SyntaxTree kid) {
		super(extendKids(existing, kid));

        nonterminal = productionType.getLhs();
    }

	protected static SyntaxTree[] extendKids(ASTList existing, SyntaxTree kid) {
        SyntaxTree[] oldKids = existing.getKids();

        SyntaxTree[] result = new SyntaxTree[oldKids.length + 1];
        for (int i = 0; i < oldKids.length; i++) {
            result[i] = oldKids[i];
        }

        result[oldKids.length] = kid;
		return result;
	}

    /** Getter for the nonterminal on the lhs of the matched production. */
    public Nonterminal getNonterminalType() {
	    return nonterminal;
    }

    public Production getProductionType() {
        return null;
    }

    /**
     * Tell this node it is now (more) ambiguous, & supply the alternative.
     * @param alternative a subtree that competes with this one.
     */
    public void ambiguity(SyntaxTreeBranch alternate) {
        assert false: "ambiguous list";
    }

    public boolean isAmbiguous() {
        return false;
    }

    public boolean isList() {
        return true;
    }
}
